#    peepdf-3 is a tool to analyse and modify PDF files
#    https://github.com/digitalsleuth/peepdf-3
#    Original Author: Jose Miguel Esparza <jesparza AT eternal-todo.com>
#    Updated for Python 3 by Corey Forman (digitalsleuth - https://github.com/digitalsleuth/peepdf-3)
#    Copyright (C) 2011-2017 Jose Miguel Esparza
#
#    This file is part of peepdf-3.
#
#        peepdf-3 is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        peepdf-3 is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with peepdf-3. If not, see <http://www.gnu.org/licenses/>.

"""
    This module contains some functions to analyse Javascript code inside the PDF file
"""

import jsbeautifier
import os
import re
import sys
import traceback

try:
    from peepdf.PDFUtils import unescapeHTMLEntities, escapeString
except ModuleNotFoundError:
    from PDFUtils import unescapeHTMLEntities, escapeString

try:
    import STPyV8

    class Global(STPyV8.JSClass):
        evalCode = ""

        def evalOverride(self, expression):
            self.evalCode += f"\n\n// New evaluated code\n{expression}"
            return

    JS_MODULE = True

except:
    JS_MODULE = False


errorsFile = "errors.txt"
newLine = os.linesep
reJSscript = "<script[^>]*?contentType\s*?=\s*?['\"]application/x-javascript['\"][^>]*?>(.*?)</script>"
preDefinedCode = "var app = this;"


def analyseJS(code, context=None, manualAnalysis=False):
    """
    Hooks the eval function and search for obfuscated elements in the Javascript code

    @param code: The Javascript code (string)
    @return: List with analysis information of the Javascript code: [JSCode,unescapedBytes,urlsFound,errors,context], where
            JSCode is a list with the several stages Javascript code,
            unescapedBytes is a list with the parameters of unescape functions,
            urlsFound is a list with the URLs found in the unescaped bytes,
            errors is a list of errors,
            context is the context of execution of the Javascript code.
    """
    errors = []
    jsCode = []
    unescapedBytes = []
    urlsFound = []

    try:
        code = unescapeHTMLEntities(code)
        scriptElements = re.findall(reJSscript, code, re.DOTALL | re.IGNORECASE)
        if scriptElements:
            code = ""
            for scriptElement in scriptElements:
                code += f"{scriptElement}{newLine * 2}"
        code = jsbeautifier.beautify(code)
        jsCode.append(code)

        if code is not None and JS_MODULE and not manualAnalysis:
            if context is None:
                context = STPyV8.JSContext(Global())
            context.enter()
            # Hooking the eval function
            context.eval("eval=evalOverride")
            # context.eval(preDefinedCode)
            while True:
                try:
                    context.eval(code)
                    evalCode = context.eval("evalCode")
                    evalCode = jsbeautifier.beautify(evalCode)
                    if evalCode != "" and evalCode != code:
                        code = evalCode
                        jsCode.append(code)
                    else:
                        break
                except:
                    error = str(sys.exc_info()[1])
                    open("jserror.log", "a").write(f"{error}{newLine}")
                    errors.append(error)
                    break

            if code != "":
                escapedVars = re.findall(
                    "(\w*?)\s*?=\s*?(unescape\((.*?)\))", code, re.DOTALL
                )
                for var in escapedVars:
                    bytes = var[2]
                    if bytes.find("+") != -1 or bytes.find("%") == -1:
                        varContent = getVarContent(code, bytes)
                        if len(varContent) > 150:
                            ret = unescape(varContent)
                            if ret[0] != -1:
                                bytes = ret[1]
                                urls = re.findall("https?://.*$", bytes, re.DOTALL)
                                if bytes not in unescapedBytes:
                                    unescapedBytes.append(bytes)
                                for url in urls:
                                    if url not in urlsFound:
                                        urlsFound.append(url)
                    else:
                        bytes = bytes[1:-1]
                        if len(bytes) > 150:
                            ret = unescape(bytes)
                            if ret[0] != -1:
                                bytes = ret[1]
                                urls = re.findall("https?://.*$", bytes, re.DOTALL)
                                if bytes not in unescapedBytes:
                                    unescapedBytes.append(bytes)
                                for url in urls:
                                    if url not in urlsFound:
                                        urlsFound.append(url)
    except:
        traceback.print_exc(file=open(errorsFile, "a"))
        errors.append("Unexpected error in the JSAnalysis module!!")
    finally:
        for js in jsCode:
            if js is None or js == "":
                jsCode.remove(js)
    return [jsCode, unescapedBytes, urlsFound, errors, context]


def getVarContent(jsCode, varContent):
    """
    Given the Javascript code and the content of a variable this method tries to obtain the real value of the variable, cleaning expressions like "a = eval; a(js_code);"

    @param jsCode: The Javascript code (string)
    @param varContent: The content of the variable (string)
    @return: A string with real value of the variable
    """
    clearBytes = ""
    varContent = varContent.replace("\n", "")
    varContent = varContent.replace("\r", "")
    varContent = varContent.replace("\t", "")
    varContent = varContent.replace(" ", "")
    parts = varContent.split("+")
    for part in parts:
        if re.match("[\"'].*?[\"']", part, re.DOTALL):
            clearBytes += part[1:-1]
        else:
            part = escapeString(part)
            varContent = re.findall(part + "\s*?=\s*?(.*?)[,;]", jsCode, re.DOTALL)
            if varContent:
                clearBytes += getVarContent(jsCode, varContent[0])
    return clearBytes


def isJavascript(content):
    """
    Given an string this method looks for typical Javscript strings and try to identify if the string contains Javascrit code or not.

    @param content: A string
    @return: A boolean, True if it seems to contain Javascript code or False in the other case
    """
    jsStrings = [
        "var ",
        ";",
        ")",
        "(",
        "function ",
        "=",
        "{",
        "}",
        "if(",
        "if (",
        "else {",
        "else{",
        "else if",
        "return",
        "while(",
        "while (",
        "for(",
        "for (",
        ",",
        "eval",
    ]
    keyStrings = [";", "(", ")"]
    stringsFound = []
    limit = 15
    minDistinctStringsFound = 5
    minRatio = 10
    results = 0
    length = len(content)
    smallScriptLength = 100

    if re.findall(reJSscript, content, re.DOTALL | re.IGNORECASE):
        return True

    contentAlphaNumeric = any(
        any(char.isalnum() and char.isascii() for char in item) for item in content
    )
    if not contentAlphaNumeric:
        return False
    for string in jsStrings:
        cont = content.count(string)
        results += cont
        if cont > 0 and string not in stringsFound:
            stringsFound.append(string)
        elif cont == 0 and string in keyStrings:
            return False

    stringsFoundAlphaNumeric = any(
        any(char.isalnum() and char.isascii() for char in item) for item in stringsFound
    )
    if not stringsFoundAlphaNumeric:
        return False
    numDistinctStringsFound = len(stringsFound)
    ratio = (results * 100.0) / length
    if (results > limit and numDistinctStringsFound >= minDistinctStringsFound) or (
        length < smallScriptLength and ratio > minRatio
    ):
        return True
    else:
        return False


def searchObfuscatedFunctions(jsCode, function):
    """
    Search for obfuscated functions in the Javascript code

    @param jsCode: The Javascript code (string)
    @param function: The function name to look for (string)
    @return: List with obfuscated functions information [functionName,functionCall,containsReturns]
    """
    obfuscatedFunctionsInfo = []
    if jsCode is not None:
        match = re.findall(
            "\W(" + function + "\s{0,5}?\((.*?)\)\s{0,5}?;)", jsCode, re.DOTALL
        )
        if match:
            for m in match:
                if re.findall("return", m[1], re.IGNORECASE):
                    obfuscatedFunctionsInfo.append([function, m, True])
                else:
                    obfuscatedFunctionsInfo.append([function, m, False])
        obfuscatedFunctions = re.findall(
            "\s*?((\w*?)\s*?=\s*?" + function + ")\s*?;", jsCode, re.DOTALL
        )
        for obfuscatedFunction in obfuscatedFunctions:
            obfuscatedElement = obfuscatedFunction[1]
            obfuscatedFunctionsInfo += searchObfuscatedFunctions(
                jsCode, obfuscatedElement
            )
    return obfuscatedFunctionsInfo


def unescape(escapedBytes, unicode=True):
    """
    This method unescapes the given string

    @param escapedBytes: A string to unescape
    @return: A tuple (status,statusContent), where statusContent is an unescaped string in case status = 0 or an error in case status = -1
    """
    # TODO: modify to accept a list of escaped strings?
    unescapedBytes = ""
    if unicode:
        unicodePadding = "\x00"
    else:
        unicodePadding = ""
    try:
        if (
            escapedBytes.lower().find("%u") != -1
            or escapedBytes.lower().find("\\u") != -1
            or escapedBytes.find("%") != -1
        ):
            if escapedBytes.lower().find("\\u") != -1:
                splitBytes = escapedBytes.split("\\")
            else:
                splitBytes = escapedBytes.split("%")
            for i in range(len(splitBytes)):
                splitByte = splitBytes[i]
                if splitByte == "":
                    continue
                if len(splitByte) > 4 and re.match(
                    "u[0-9a-f]{4}", splitByte[:5], re.IGNORECASE
                ):
                    unescapedBytes += chr(int(splitByte[3] + splitByte[4], 16)) + chr(
                        int(splitByte[1] + splitByte[2], 16)
                    )
                    if len(splitByte) > 5:
                        for j in range(5, len(splitByte)):
                            unescapedBytes += splitByte[j] + unicodePadding
                elif len(splitByte) > 1 and re.match(
                    "[0-9a-f]{2}", splitByte[:2], re.IGNORECASE
                ):
                    unescapedBytes += (
                        chr(int(splitByte[0] + splitByte[1], 16)) + unicodePadding
                    )
                    if len(splitByte) > 2:
                        for j in range(2, len(splitByte)):
                            unescapedBytes += splitByte[j] + unicodePadding
                else:
                    if i != 0:
                        unescapedBytes += "%" + unicodePadding
                    for j in range(len(splitByte)):
                        unescapedBytes += splitByte[j] + unicodePadding
        else:
            unescapedBytes = escapedBytes
    except:
        return (-1, "[!] Error while unescaping the bytes")
    return (0, unescapedBytes)



#    peepdf-3 is a tool to analyse and modify PDF files
#    https://github.com/digitalsleuth/peepdf-3
#    Original Author: Jose Miguel Esparza <jesparza AT eternal-todo.com>
#    Updated for Python 3 by Corey Forman (digitalsleuth - https://github.com/digitalsleuth/peepdf-3)
#    Copyright (C) 2011-2017 Jose Miguel Esparza
#
#    This file is part of peepdf-3.
#
#        peepdf-3 is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        peepdf-3 is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with peepdf-3. If not, see <http://www.gnu.org/licenses/>.

from . import PDFConsole, PDFCore, PDFCrypto, PDFFilters, PDFUtils, PDFVulns, PDFEnDec, JSAnalysis



#    peepdf-3 is a tool to analyse and modify PDF files
#    https://github.com/digitalsleuth/peepdf-3
#    Original Author: Jose Miguel Esparza <jesparza AT eternal-todo.com>
#    Updated for Python 3 by Corey Forman (digitalsleuth - https://github.com/digitalsleuth/peepdf-3)
#    Copyright (C) 2011-2017 Jose Miguel Esparza
#
#    This file is part of peepdf-3.
#
#        peepdf-3 is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        peepdf-3 is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with peepdf-3. If not, see <http://www.gnu.org/licenses/>.

"""
    Module to manage cryptographic operations with PDF files
"""

import hashlib, struct, random, warnings, sys
from aespython import key_expander, aes_cipher, cbc_mode

warnings.filterwarnings("ignore")

paddingString = b"\x28\xBF\x4E\x5E\x4E\x75\x8A\x41\x64\x00\x4E\x56\xFF\xFA\x01\x08\x2E\x2E\x00\xB6\xD0\x68\x3E\x80\x2F\x0C\xA9\xFE\x64\x53\x69\x7A"


def computeEncryptionKey(
    password,
    dictOwnerPass,
    dictUserPass,
    dictOE,
    dictUE,
    fileID,
    pElement,
    dictKeyLength=128,
    revision=3,
    encryptMetadata=False,
    passwordType=None,
):
    """
    Compute an encryption key to encrypt/decrypt the PDF file

    @param password: The password entered by the user
    @param dictOwnerPass: The owner password from the standard security handler dictionary
    @param dictUserPass: The user password from the standard security handler dictionary
    @param dictOE: The owner encrypted string from the standard security handler dictionary
    @param dictUE:The user encrypted string from the standard security handler dictionary
    @param fileID: The /ID element in the trailer dictionary of the PDF file
    @param pElement: The /P element of the Encryption dictionary
    @param dictKeyLength: The length of the key
    @param revision: The algorithm revision
    @param encryptMetadata: A boolean extracted from the standard security handler dictionary to specify if it's necessary to encrypt the document metadata or not
    @param passwordType: It specifies the given password type. It can be 'USER', 'OWNER' or None.
    @return: A tuple (status,statusContent), where statusContent is the encryption key in case status = 0 or an error message in case status = -1
    """
    try:
        if revision != 5:
            keyLength = dictKeyLength / 8
            lenPass = len(password)
            if lenPass > 32:
                password = password[:32]
            elif lenPass < 32:
                password += paddingString[: 32 - lenPass]
            md5input = (
                password + dictOwnerPass + struct.pack("<i", int(pElement)) + fileID
            )
            if revision > 3 and not encryptMetadata:
                md5input += b"\xFF" * 4
            key = hashlib.md5(md5input).digest()
            if revision > 2:
                counter = 0
                while counter < 50:
                    key = hashlib.md5(key[:keyLength]).digest()
                    counter += 1
                key = key[:keyLength]
            elif revision == 2:
                key = key[:5]
            return (0, key)
        else:
            if passwordType == "USER":
                password = password[:127]
                kSalt = dictUserPass[40:48]
                intermediateKey = hashlib.sha256(password + kSalt).digest()
                ret = decryptData(b"\0" * 16 + dictUE, intermediateKey)
            elif passwordType == "OWNER":
                password = password[:127]
                kSalt = dictOwnerPass[40:48]
                intermediateKey = hashlib.sha256(
                    password + kSalt + dictUserPass
                ).digest()
                ret = decryptData(b"\0" * 16 + dictOE, intermediateKey)
            return ret
    except:
        return (
            -1,
            f"ComputeEncryptionKey error: {str(sys.exc_info()[0])} {str(sys.exc_info()[1])}",
        )


def computeObjectKey(id, generationNum, encryptionKey, keyLengthBytes, algorithm="RC4"):
    """
    Compute the key necessary to encrypt each object, depending on the id and generation number. Only necessary with /V < 5.

    @param id: The object id
    @param generationNum: The generation number of the object
    @param encryptionKey: The encryption key
    @param keyLengthBytes: The length of the encryption key in bytes
    @param algorithm: The algorithm used in the encryption/decryption process
    @return A tuple (status,statusContent), where statusContent is the computed key in case status = 0 or an error message in case status = -1
    """
    try:
        key = (
            encryptionKey
            + struct.pack("<i", id)[:3]
            + struct.pack("<i", generationNum)[:2]
        )
        if algorithm == "AES":
            key += "\x73\x41\x6C\x54"  # sAlT
        key = hashlib.md5(key).digest()
        if keyLengthBytes + 5 < 16:
            key = key[: keyLengthBytes + 5]
        else:
            key = key[:16]
        # AES: block size = 16 bytes, initialization vector (16 bytes), random, first bytes encrypted string
        return (0, key)
    except:
        return (
            -1,
            f"ComputeObjectKey error: {str(sys.exc_info()[0])} {str(sys.exc_info()[1])}",
        )


def computeOwnerPass(ownerPassString, userPassString, keyLength=128, revision=3):
    """
    Compute the owner password necessary to compute the encryption key of the PDF file

    @param ownerPassString: The owner password entered by the user
    @param userPassString: The user password entered by the user
    @param keyLength: The length of the key
    @param revision: The algorithm revision
    @return A tuple (status,statusContent), where statusContent is the computed password in case status = 0 or an error message in case status = -1
    """
    try:
        # TODO: revision 5
        keyLength = keyLength / 8
        lenPass = len(ownerPassString)
        if lenPass > 32:
            ownerPassString = ownerPassString[:32]
        elif lenPass < 32:
            ownerPassString += paddingString[: 32 - lenPass]
        rc4Key = hashlib.md5(ownerPassString).digest()
        if revision > 2:
            counter = 0
            while counter < 50:
                rc4Key = hashlib.md5(rc4Key).digest()
                counter += 1
        rc4Key = rc4Key[:keyLength]
        lenPass = len(userPassString)
        if lenPass > 32:
            userPassString = userPassString[:32]
        elif lenPass < 32:
            userPassString += paddingString[: 32 - lenPass]
        ownerPass = RC4(userPassString, rc4Key)
        if revision > 2:
            counter = 1
            while counter <= 19:
                newKey = ""
                for i in range(len(rc4Key)):
                    newKey += chr(ord(rc4Key[i]) ^ counter)
                ownerPass = RC4(ownerPass, newKey)
                counter += 1
        return (0, ownerPass)
    except:
        return (
            -1,
            f"ComputeOwnerPass error: {str(sys.exc_info()[0])} {str(sys.exc_info()[1])}",
        )


def computeUserPass(
    userPassString,
    dictO,
    fileID,
    pElement,
    keyLength=128,
    revision=3,
    encryptMetadata=False,
):
    """
    Compute the user password of the PDF file

    @param userPassString: The user password entered by the user
    @param ownerPass: The computed owner password
    @param fileID: The /ID element in the trailer dictionary of the PDF file
    @param pElement: The /P element of the /Encryption dictionary
    @param keyLength: The length of the key
    @param revision: The algorithm revision
    @param encryptMetadata: A boolean extracted from the standard security handler dictionary to specify if it's necessary to encrypt the document metadata or not
    @return: A tuple (status,statusContent), where statusContent is the computed password in case status = 0 or an error message in case status = -1
    """
    # TODO: revision 5
    userPass = ""
    dictU = ""
    dictOE = ""
    dictUE = ""
    ret = computeEncryptionKey(
        userPassString,
        dictO,
        dictU,
        dictOE,
        dictUE,
        fileID,
        pElement,
        keyLength,
        revision,
        encryptMetadata,
    )
    if ret[0] != -1:
        rc4Key = ret[1]
    else:
        return ret
    try:
        if revision == 2:
            userPass = RC4(paddingString, rc4Key)
        elif revision > 2:
            counter = 1
            md5Input = paddingString + fileID
            hashResult = hashlib.md5(md5Input).digest()
            userPass = RC4(hashResult, rc4Key)
            while counter <= 19:
                newKey = ""
                for i in range(len(rc4Key)):
                    newKey += chr(ord(rc4Key[i]) ^ counter)
                userPass = RC4(userPass, newKey)
                counter += 1
            counter = 0
            while counter < 16:
                userPass += chr(random.randint(32, 255))
                counter += 1
        else:
            # This should not be possible or the PDF specification does not say anything about it
            return (-1, f"ComputeUserPass error: revision number is < 2 ({revision})")
        return (0, userPass)
    except:
        return (
            -1,
            f"ComputeUserPass error: {str(sys.exc_info()[0])} {str(sys.exc_info()[1])}",
        )


def isUserPass(password, computedUserPass, dictU, revision):
    """
    Checks if the given password is the User password of the file

    @param password: The given password or the empty password
    @param computedUserPass: The computed user password of the file
    @param dictU: The /U element of the /Encrypt dictionary
    @param revision: The number of revision of the standard security handler
    @return The boolean telling if the given password is the user password or not
    """

    if revision == 5:
        vSalt = dictU[32:40]
        inputHash = hashlib.sha256(password + vSalt).digest()
        if inputHash == dictU[:32]:
            return True
        else:
            return False
    elif revision == 3 or revision == 4:
        if computedUserPass[:16] == dictU[:16]:
            return True
        else:
            return False
    elif revision < 3:
        if computedUserPass == dictU:
            return True
        else:
            return False


def isOwnerPass(password, dictO, dictU, computedUserPass, keyLength, revision):
    """
    Checks if the given password is the owner password of the file

    @param password: The given password or the empty password
    @param dictO: The /O element of the /Encrypt dictionary
    @param dictU: The /U element of the /Encrypt dictionary
    @param computedUserPass: The computed user password of the file
    @param keyLength: The length of the key
    @param revision: The algorithm revision
    @return The boolean telling if the given password is the owner password or not
    """
    if revision == 5:
        vSalt = dictO[32:40]
        inputHash = hashlib.sha256(password + vSalt + dictU).digest()
        if inputHash == dictO[:32]:
            return True
        else:
            return False
    else:
        keyLength = int(keyLength / 8)
        lenPass = len(password)
        if lenPass > 32:
            password = password[:32]
        elif lenPass < 32:
            password += paddingString[: 32 - lenPass]
        rc4Key = hashlib.md5(password).digest()
        if revision > 2:
            counter = 0
            while counter < 50:
                rc4Key = hashlib.md5(rc4Key).digest()
                counter += 1
        rc4Key = rc4Key[:keyLength]
        if revision == 2:
            userPass = RC4(dictO, rc4Key)
        elif revision > 2:
            counter = 19
            while counter >= 0:
                newKey = ""
                for i in range(len(rc4Key)):
                    newKey += chr(ord(rc4Key[i]) ^ counter)
                dictO = RC4(dictO, newKey)
                counter -= 1
            userPass = dictO
        else:
            # Is it possible??
            userPass = ""
        return isUserPass(userPass, computedUserPass, dictU, revision)


def RC4(data, key):
    """
    RC4 implementation

    @param data: Bytes to be encrypyed/decrypted
    @param key: Key used for the algorithm
    @return: The encrypted/decrypted bytes
    """
    y = 0
    hash = {}
    box = {}
    ret = ""
    keyLength = len(key)
    dataLength = len(data)

    # Initialization
    for x in range(256):
        hash[x] = key[x % keyLength]
        box[x] = x
    for x in range(256):
        y = (y + int(box[x]) + int(hash[x])) % 256
        tmp = box[x]
        box[x] = box[y]
        box[y] = tmp

    z = y = 0
    for x in range(0, dataLength):
        z = (z + 1) % 256
        y = (y + box[z]) % 256
        tmp = box[z]
        box[z] = box[y]
        box[y] = tmp
        k = box[((box[z] + box[y]) % 256)]
        ret += chr(data[x] ^ k)
    if isinstance(ret, str):
        ret = ret.encode('latin-1')
    return ret

def xor(bytes, key):
    """
    Simple XOR implementation
    Author: Evan Fosmark (http://www.evanfosmark.com/2008/06/xor-encryption-with-python/)
    @param bytes: Bytes to be xored
    @param key: Key used for the operation, it's cycled.
    @return: The xored bytes
    """
    key = cycle(key)
    return "".join(chr(ord(x) ^ ord(y)) for (x, y) in zip(bytes, key))

def decryptData(data, password=None, keyLength=None, mode="CBC"):
    """
    Created from the demonstration of the pythonaes package.
    
    Copyright (c) 2010, Adam Newman http://www.caller9.com/
    Licensed under the MIT license http://www.opensource.org/licenses/mit-license.php
    """
    decryptedData = ""
    if keyLength == None:
        keyLength = len(password) * 8
    if keyLength not in [128, 192, 256]:
        return (-1, "Bad length key in AES decryption process")
    try:
        iv = list(map(ord, data[:16]))
    except:
        iv = list(data[:16])
    try:
        key = list(map(ord, password))
    except:
        key = list(password)
    data = data[16:]
    if len(data) % 16 != 0:
        data = data[: -(len(data) % 16)]
    keyExpander = key_expander.KeyExpander(keyLength)
    expandedKey = keyExpander.expand(key)
    aesCipher = aes_cipher.AESCipher(expandedKey)
    if mode == "CBC":
        aesMode = cbc_mode.CBCMode(aesCipher, 16)
    aesMode.set_iv(iv)
    for i in range(0, len(data), 16):
        try:
            ciphertext = list(map(ord, data[i : i + 16]))
        except:
            ciphertext = list(data[i : i + 16])
        decryptedBytes = aesMode.decrypt_block(ciphertext)
        for byte in decryptedBytes:
            decryptedData += chr(byte)
    return (0, decryptedData)


#    peepdf-3 is a tool to analyse and modify PDF files
#    https://github.com/digitalsleuth/peepdf-3
#    Original Author: Jose Miguel Esparza <jesparza AT eternal-todo.com>
#    Updated for Python 3 by Corey Forman (digitalsleuth - https://github.com/digitalsleuth/peepdf-3)
#    Copyright (C) 2011-2017 Jose Miguel Esparza
#
#    This file is part of peepdf-3.
#
#        peepdf-3 is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        peepdf-3 is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with peepdf-3. If not, see <http://www.gnu.org/licenses/>.

# Some code has been reused and modified from the original by Mathieu Fenniak:
# Parameters management in Flate and LZW algorithms, asciiHexDecode and ascii85Decode
#
# Copyright (c) 2006, Mathieu Fenniak
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


"""
    Module to manage encoding/decoding in PDF files
"""

import sys, zlib, struct
from binascii import hexlify, unhexlify

try:
    from peepdf.PDFEnDec import *
    from peepdf.PDFUtils import getNumsFromBytes, getBytesFromBits, getBitsFromNum
except ModuleNotFoundError:
    from PDFEnDec import *
    from PDFUtils import getNumsFromBytes, getBytesFromBits, getBitsFromNum


def decodeStream(stream, filter, parameters={}):
    """
    Decode the given stream

    @param stream: Stream to be decoded (string)
    @param filter: Filter to apply to decode the stream
    @param parameters: List of PDFObjects containing the parameters for the filter
    @return: A tuple (status,statusContent), where statusContent is the decoded stream in case status = 0 or an error in case status = -1
    """
    if filter == "/ASCIIHexDecode" or filter == "/AHx":
        ret = asciiHexDecode(stream)
    elif filter == "/ASCII85Decode" or filter == "/A85":
        ret = ascii85Decode(stream)
    elif filter == "/LZWDecode" or filter == "/LZW":
        ret = lzwDecode(stream, parameters)
    elif filter == "/FlateDecode" or filter == "/Fl":
        ret = flateDecode(stream, parameters)
    elif filter == "/RunLengthDecode" or filter == "/RL":
        ret = runLengthDecode(stream)
    elif filter == "/CCITTFaxDecode" or filter == "/CCF":
        ret = ccittFaxDecode(stream, parameters)
    elif filter == "/JBIG2Decode":
        ret = jbig2Decode(stream, parameters)
    elif filter == "/DCTDecode" or filter == "/DCT":
        ret = dctDecode(stream, parameters)
    elif filter == "/JPXDecode":
        ret = jpxDecode(stream)
    elif filter == "/Crypt":
        ret = crypt(stream, parameters)
    else:
        ret = (-1, f'Unknown filter "{filter}"')
    return ret


def encodeStream(stream, filter, parameters={}):
    """
    Encode the given stream

    @param stream: Stream to be decoded (string)
    @param filter: Filter to apply to decode the stream
    @param parameters: List of PDFObjects containing the parameters for the filter
    @return: A tuple (status,statusContent), where statusContent is the encoded stream in case status = 0 or an error in case status = -1
    """
    if filter == "/ASCIIHexDecode":
        ret = asciiHexEncode(stream)
    elif filter == "/ASCII85Decode":
        ret = ascii85Encode(stream)
    elif filter == "/LZWDecode":
        ret = lzwEncode(stream, parameters)
    elif filter == "/FlateDecode":
        ret = flateEncode(stream, parameters)
    elif filter == "/RunLengthDecode":
        ret = runLengthEncode(stream)
    elif filter == "/CCITTFaxDecode":
        ret = ccittFaxEncode(stream, parameters)
    elif filter == "/JBIG2Decode":
        ret = jbig2Encode(stream, parameters)
    elif filter == "/DCTDecode":
        ret = dctEncode(stream, parameters)
    elif filter == "/JPXDecode":
        ret = jpxEncode(stream)
    elif filter == "/Crypt":
        ret = crypt(stream, parameters)
    else:
        ret = (-1, f'Unknown filter "{filter}"')
    return ret


'''
The ascii85Decode code is part of pdfminer (http://pypi.python.org/pypi/pdfminer/)

Copyright (c) 2004-2010 Yusuke Shinyama <yusuke at cs dot nyu dot edu>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

    """
    In ASCII85 encoding, every four bytes are encoded with five ASCII
    letters, using 85 different types of characters (as 256**4 < 85**5).
    When the length of the original bytes is not a multiple of 4, a special
    rule is used for round up.

    The Adobe's ASCII85 implementation is slightly different from
    its original in handling the last characters.

    The sample string is taken from:
      http://en.wikipedia.org/w/index.php?title=Ascii85

    >>> ascii85decode('9jqo^BlbD-BleB1DJ+*+F(f,q')
    'Man is distinguished'
    >>> ascii85decode('E,9)oF*2M7/c~>')
    'pleasure.'
    """

'''


def ascii85Decode(stream):
    """
    Method to decode streams using ASCII85

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decoded PDF stream in case status = 0 or an error in case status = -1
    """
    n = b = 0
    decodedStream = ""
    try:
        for c in stream:
            if "!" <= c and c <= "u":
                n += 1
                b = b * 85 + (ord(c) - 33)
                if n == 5:
                    decodedStream += struct.pack(">L", b)
                    n = b = 0
            elif c == "z":
                assert n == 0
                decodedStream += "\0\0\0\0"
            elif c == "~":
                if n:
                    for _ in range(5 - n):
                        b = b * 85 + 84
                    decodedStream += struct.pack(">L", b)[: n - 1]
                break
    except:
        return (-1, "Unspecified error")
    return (0, decodedStream)


def ascii85Encode(stream):
    """
    Method to encode streams using ASCII85 (NOT SUPPORTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encoded PDF stream in case status = 0 or an error in case status = -1
    TODO: use base64 a85 encode and decode
    """

    return (-1, "Ascii85Encode not supported yet")


def asciiHexDecode(stream):
    """
    Method to decode streams using hexadecimal encoding

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decoded PDF stream in case status = 0 or an error in case status = -1
    """
    eod = ">"
    decodedStream = ""
    char = ""
    index = 0
    while index < len(stream):
        c = stream[index]
        if c == eod:
            if len(decodedStream) % 2 != 0:
                char += "0"
                try:
                    decodedStream += chr(int(char, 16))
                except:
                    return (-1, "Error in hexadecimal conversion")
            break
        elif c.isspace():
            index += 1
            continue
        char += c
        if len(char) == 2:
            try:
                decodedStream += chr(int(char, 16))
            except:
                return (-1, "Error in hexadecimal conversion")
            char = ""
        index += 1
    return (0, decodedStream)


def asciiHexEncode(stream):
    """
    Method to encode streams using hexadecimal encoding

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encoded PDF stream in case status = 0 or an error in case status = -1
    Confirm whether or not Unicode correction is needed here.
    """
    try:
        encodedStream = hexlify(stream.encode()).decode("latin-1")
    except:
        return (-1, "Error in hexadecimal conversion")
    return (0, encodedStream)


def flateDecode(stream, parameters):
    """
    Method to decode streams using the Flate algorithm

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decoded PDF stream in case status = 0 or an error in case status = -1
    """
    decodedStream = ""
    try:
        if not isinstance(stream, bytes):
            decodedStream = (zlib.decompress(stream.encode("latin-1"))).decode(
                "latin-1"
            )
        else:
            decodedStream = zlib.decompress(stream).decode("latin-1")
    except:
        return (-1, "Error decompressing string")

    if parameters is None or parameters == {}:
        return (0, decodedStream)
    else:
        if "/Predictor" in parameters:
            predictor = parameters["/Predictor"].getRawValue()
        else:
            predictor = 1
        # Columns = number of samples per row
        if "/Columns" in parameters:
            columns = parameters["/Columns"].getRawValue()
        else:
            columns = 1
        # Colors = number of components per sample
        if "/Colors" in parameters:
            colors = parameters["/Colors"].getRawValue()
            if colors < 1:
                colors = 1
        else:
            colors = 1
        # BitsPerComponent: number of bits per color component
        if "/BitsPerComponent" in parameters:
            bits = parameters["/BitsPerComponent"].getRawValue()
            if bits not in [1, 2, 4, 8, 16]:
                bits = 8
        else:
            bits = 8
        if predictor is not None and predictor != 1:
            ret = post_prediction(decodedStream, predictor, columns, colors, bits)
            return ret
        else:
            return (0, decodedStream)


def flateEncode(stream, parameters):
    """
    Method to encode streams using the Flate algorithm

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encoded PDF stream in case status = 0 or an error in case status = -1
    """
    encodedStream = ""
    if parameters is None or parameters == {}:
        try:
            return (0, zlib.compress(stream.encode("latin-1")).decode("latin-1"))
        except:
            return (-1, "Error compressing string")
    else:
        if "/Predictor" in parameters:
            predictor = parameters["/Predictor"].getRawValue()
        else:
            predictor = 1
        # Columns = number of samples per row
        if "/Columns" in parameters:
            columns = parameters["/Columns"].getRawValue()
        else:
            columns = 1
        # Colors = number of components per sample
        if "/Colors" in parameters:
            colors = parameters["/Colors"].getRawValue()
            if colors < 1:
                colors = 1
        else:
            colors = 1
        # BitsPerComponent: number of bits per color component
        if "/BitsPerComponent" in parameters:
            bits = parameters["/BitsPerComponent"].getRawValue()
            if bits not in [1, 2, 4, 8, 16]:
                bits = 8
        else:
            bits = 8
        if predictor is not None and predictor != 1:
            ret = pre_prediction(stream, predictor, columns, colors, bits)
            if ret[0] == -1:
                return ret
            output = ret[1]
        else:
            output = stream
        try:
            return (0, zlib.compress(output.encode("latin-1")).decode("latin-1"))
        except:
            return (-1, "Error compressing string")


def lzwDecode(stream, parameters):
    """
    Method to decode streams using the LZW algorithm

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decoded PDF stream in case status = 0 or an error in case status = -1
    """
    decodedStream = ""
    try:
        decodedStream = lzw.lzwdecode(stream)
    except:
        return (-1, "Error decompressing string")

    if parameters is None or parameters == {}:
        return (0, decodedStream)
    else:
        if "/Predictor" in parameters:
            predictor = parameters["/Predictor"].getRawValue()
        else:
            predictor = 1
        # Columns = number of samples per row
        if "/Columns" in parameters:
            columns = parameters["/Columns"].getRawValue()
        else:
            columns = 1
        # Colors = number of components per sample
        if "/Colors" in parameters:
            colors = parameters["/Colors"].getRawValue()
            if colors < 1:
                colors = 1
        else:
            colors = 1
        # BitsPerComponent: number of bits per color component
        if "/BitsPerComponent" in parameters:
            bits = parameters["/BitsPerComponent"].getRawValue()
            if bits not in [1, 2, 4, 8, 16]:
                bits = 8
        else:
            bits = 8
        if "/EarlyChange" in parameters:
            earlyChange = parameters["/EarlyChange"].getRawValue()
        else:
            earlyChange = 1
        if predictor is not None and predictor != 1:
            ret = post_prediction(decodedStream, predictor, columns, colors, bits)
            return ret
        else:
            return (0, decodedStream)


def lzwEncode(stream, parameters):
    """
    Method to encode streams using the LZW algorithm

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encoded PDF stream in case status = 0 or an error in case status = -1
    """
    encodedStream = ""
    if parameters is None or parameters == {}:
        try:
            generator = lzw.compress(stream)
            for c in generator:
                encodedStream += c
            return (0, encodedStream)
        except:
            return (-1, "Error compressing string")
    else:
        if "/Predictor" in parameters:
            predictor = parameters["/Predictor"].getRawValue()
        else:
            predictor = 1
        # Columns = number of samples per row
        if "/Columns" in parameters:
            columns = parameters["/Columns"].getRawValue()
        else:
            columns = 1
        # Colors = number of components per sample
        if "/Colors" in parameters:
            colors = parameters["/Colors"].getRawValue()
            if colors < 1:
                colors = 1
        else:
            colors = 1
        # BitsPerComponent: number of bits per color component
        if "/BitsPerComponent" in parameters:
            bits = parameters["/BitsPerComponent"].getRawValue()
            if bits not in [1, 2, 4, 8, 16]:
                bits = 8
        else:
            bits = 8
        if "/EarlyChange" in parameters:
            earlyChange = parameters["/EarlyChange"].getRawValue()
        else:
            earlyChange = 1
        if predictor is not None and predictor != 1:
            ret = pre_prediction(stream, predictor, columns, colors, bits)
            if ret[0] == -1:
                return ret
            output = ret[1]
        else:
            output = stream
        try:
            generator = lzw.compress(output)
            for c in generator:
                encodedStream += c
            return (0, encodedStream)
        except:
            return (-1, "Error decompressing string")


def pre_prediction(stream, predictor, columns, colors, bits):
    """
    Predictor function to make the stream more predictable and improve compression (PDF Specification)

    @param stream: The stream to be modified
    @param predictor: The type of predictor to apply
    @param columns: Number of samples per row
    @param colors: Number of colors per sample
    @param bits: Number of bits per color
    @return: A tuple (status,statusContent), where statusContent is the modified stream in case status = 0 or an error in case status = -1
    """

    output = ""
    # TODO: TIFF and more PNG predictions

    # PNG prediction
    if predictor >= 10 and predictor <= 15:
        # PNG prediction can vary from row to row
        for row in range(len(stream) / columns):
            rowdata = [ord(x) for x in stream[(row * columns) : ((row + 1) * columns)]]
            filterByte = predictor - 10
            rowdata = [filterByte] + rowdata
            if filterByte == 0:
                pass
            elif filterByte == 1:
                for i in range(len(rowdata) - 1, 1, -1):
                    if rowdata[i] < rowdata[i - 1]:
                        rowdata[i] = rowdata[i] + 256 - rowdata[i - 1]
                    else:
                        rowdata[i] = rowdata[i] - rowdata[i - 1]
            elif filterByte == 2:
                (-1, "Unsupported predictor")
            else:
                return (-1, "Unsupported predictor")
            output += "".join([chr(x) for x in rowdata])
        return (0, output)
    else:
        return (-1, "Unsupported predictor")


def post_prediction(decodedStream, predictor, columns, colors, bits):
    """
    Predictor function to obtain the real stream, removing the prediction (PDF Specification)

    @param decodedStream: The decoded stream to be modified
    @param predictor: The type of predictor to apply
    @param columns: Number of samples per row
    @param colors: Number of colors per sample
    @param bits: Number of bits per color
    @return: A tuple (status,statusContent), where statusContent is the modified decoded stream in case status = 0 or an error in case status = -1
    """

    output = ""
    bytesPerRow = (colors * bits * columns + 7) / 8

    # TIFF - 2
    # http://www.gnupdf.org/PNG_and_TIFF_Predictors_Filter#TIFF
    if predictor == 2:
        numRows = len(decodedStream) / bytesPerRow
        bitmask = 2**bits - 1
        outputBitsStream = ""
        for rowIndex in range(numRows):
            row = decodedStream[
                rowIndex * bytesPerRow : rowIndex * bytesPerRow + bytesPerRow
            ]
            ret, colorNums = getNumsFromBytes(row, bits)
            if ret == -1:
                return (ret, colorNums)
            pixel = [0 for x in range(colors)]
            for i in range(columns):
                for j in range(colors):
                    diffPixel = colorNums[i + j]
                    pixel[j] = (pixel[j] + diffPixel) & bitmask
                    ret, outputBits = getBitsFromNum(pixel[j], bits)
                    if ret == -1:
                        return (ret, outputBits)
                    outputBitsStream += outputBits
        output = getBytesFromBits(outputBitsStream)
        return output
    # PNG prediction
    # http://www.libpng.org/pub/png/spec/1.2/PNG-Filters.html
    # http://www.gnupdf.org/PNG_and_TIFF_Predictors_Filter#TIFF
    elif predictor >= 10 and predictor <= 15:
        bytesPerRow += 1
        numRows = int((len(decodedStream) + bytesPerRow - 1) / bytesPerRow)
        numSamplesPerRow = columns + 1
        bytesPerSample = (colors * bits + 7) / 8
        upRowdata = (0,) * numSamplesPerRow
        for row in range(numRows):
            rowdata = [
                ord(x)
                for x in decodedStream[
                    (row * int(bytesPerRow)) : ((row + 1) * int(bytesPerRow))
                ]
            ]
            # PNG prediction can vary from row to row
            filterByte = rowdata[0]
            rowdata[0] = 0

            if filterByte == 0:
                # None
                pass
            elif filterByte == 1:
                # Sub - 11
                for i in range(1, numSamplesPerRow):
                    if i < bytesPerSample:
                        prevSample = 0
                    else:
                        prevSample = rowdata[i - bytesPerSample]
                    rowdata[i] = (rowdata[i] + prevSample) % 256
            elif filterByte == 2:
                # Up - 12
                for i in range(1, numSamplesPerRow):
                    upSample = upRowdata[i]
                    rowdata[i] = (rowdata[i] + upSample) % 256
            elif filterByte == 3:
                # Average - 13
                for i in range(1, numSamplesPerRow):
                    upSample = upRowdata[i]
                    if i < bytesPerSample:
                        prevSample = 0
                    else:
                        prevSample = rowdata[i - bytesPerSample]
                    rowdata[i] = (rowdata[i] + ((prevSample + upSample) / 2)) % 256
            elif filterByte == 4:
                # Paeth - 14
                for i in range(1, numSamplesPerRow):
                    upSample = upRowdata[i]
                    if i < bytesPerSample:
                        prevSample = 0
                        upPrevSample = 0
                    else:
                        prevSample = rowdata[i - bytesPerSample]
                        upPrevSample = upRowdata[i - bytesPerSample]
                    p = prevSample + upSample - upPrevSample
                    pa = abs(p - prevSample)
                    pb = abs(p - upSample)
                    pc = abs(p - upPrevSample)
                    if pa <= pb and pa <= pc:
                        nearest = prevSample
                    elif pb <= pc:
                        nearest = upSample
                    else:
                        nearest = upPrevSample
                    rowdata[i] = (rowdata[i] + nearest) % 256
            else:
                # Optimum - 15
                # return (-1,'Unsupported predictor')
                pass
            upRowdata = rowdata
            output += "".join([chr(x) for x in rowdata[1:]])
        return (0, output)
    else:
        return (-1, "Wrong value for predictor")


def runLengthDecode(stream):
    """
    Method to decode streams using the Run-Length algorithm

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decoded PDF stream in case status = 0 or an error in case status = -1
    """
    decodedStream = ""
    index = 0
    try:
        while index < len(stream):
            length = ord(stream[index])
            if length >= 0 and length < 128:
                decodedStream += stream[index + 1 : index + length + 2]
                index += length + 2
            elif length > 128 and length < 256:
                decodedStream += stream[index + 1] * (257 - length)
                index += 2
            else:
                break
    except:
        return (-1, "Error decoding string")
    return (0, decodedStream)


def runLengthEncode(stream):
    """
    Method to encode streams using the Run-Length algorithm (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encoded PDF stream in case status = 0 or an error in case status = -1
    """
    encodedStream = ""
    return (-1, "RunLengthEncode not supported yet")


def ccittFaxDecode(stream, parameters):
    """
    Method to decode streams using the CCITT facsimile standard

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decoded PDF stream in case status = 0 or an error in case status = -1
    """
    decodedStream = ""

    if parameters is None or parameters == {}:
        try:
            decodedStream = CCITTFax().decode(stream)
            return (0, decodedStream)
        except:
            return (-1, "Error decompressing string")
    else:
        # K = A code identifying the encoding scheme used
        if "/K" in parameters:
            k = parameters["/K"].getRawValue()
            if type(k) != int:
                k = 0
            else:
                if k != 0:
                    # Only supported "Group 3, 1-D" encoding (Pure one-dimensional encoding)
                    return (-1, "CCITT encoding scheme not supported")
        else:
            k = 0
        # EndOfLine = A flag indicating whether end-of-line bit patterns are required to be present in the encoding.
        if "/EndOfLine" in parameters:
            eol = parameters["/EndOfLine"].getRawValue()
            if eol == "true":
                eol = True
            else:
                eol = False
        else:
            eol = False
        # EncodedByteAlign = A flag indicating whether the filter expects extra 0 bits before each encoded line so that the line begins on a byte boundary
        if "/EncodedByteAlign" in parameters:
            byteAlign = parameters["/EncodedByteAlign"].getRawValue()
            if byteAlign == "true":
                byteAlign = True
            else:
                byteAlign = False
        else:
            byteAlign = False
        # Columns = The width of the image in pixels.
        if "/Columns" in parameters:
            columns = parameters["/Columns"].getRawValue()
            if type(columns) != int:
                columns = 1728
        else:
            columns = 1728
        # Rows = The height of the image in scan lines.
        if "/Rows" in parameters:
            rows = parameters["/Rows"].getRawValue()
            if type(rows) != int:
                rows = 0
        else:
            rows = 0
        # EndOfBlock = number of samples per row
        if "/EndOfBlock" in parameters:
            eob = parameters["/EndOfBlock"].getRawValue()
            if eob == "false":
                eob = False
            else:
                eob = True
        else:
            eob = True
        # BlackIs1 = A flag indicating whether 1 bits are to be interpreted as black pixels and 0 bits as white pixels
        if "/BlackIs1" in parameters:
            blackIs1 = parameters["/BlackIs1"].getRawValue()
            if blackIs1 == "true":
                blackIs1 = True
            else:
                blackIs1 = False
        else:
            blackIs1 = False
        # DamagedRowsBeforeError = The number of damaged rows of data to be tolerated before an error occurs
        if "/DamagedRowsBeforeError" in parameters:
            damagedRowsBeforeError = parameters["/DamagedRowsBeforeError"].getRawValue()
        else:
            damagedRowsBeforeError = 0
        try:
            decodedStream = CCITTFax().decode(
                stream,
                k,
                eol,
                byteAlign,
                columns,
                rows,
                eob,
                blackIs1,
                damagedRowsBeforeError,
            )
            return (0, decodedStream)
        except:
            return (-1, "Error decompressing string")


def ccittFaxEncode(stream, parameters):
    """
    Method to encode streams using the CCITT facsimile standard (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encoded PDF stream in case status = 0 or an error in case status = -1
    """
    encodedStream = ""
    return (-1, "CcittFaxEncode not supported yet")


def crypt(stream, parameters):
    """
    Method to encrypt streams using a PDF security handler (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encrypted PDF stream in case status = 0 or an error in case status = -1
    """
    decodedStream = ""
    if parameters is None or parameters == {}:
        return (0, stream)
    else:
        if not "/Name" in parameters or parameters["/Name"] is None:
            return (0, stream)
        else:
            cryptFilterName = parameters["/Name"].getValue()
            if cryptFilterName == "Identity":
                return (0, stream)
            else:
                # TODO: algorithm is cryptFilterName, specified in the /CF dictionary
                return (-1, "Crypt not supported yet")


def decrypt(stream, parameters):
    """
    Method to decrypt streams using a PDF security handler (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decrypted PDF stream in case status = 0 or an error in case status = -1
    """
    encodedStream = ""
    if parameters is None or parameters == {}:
        return (0, stream)
    else:
        if not "/Name" in parameters or parameters["/Name"] is None:
            return (0, stream)
        else:
            cryptFilterName = parameters["/Name"].getValue()
            if cryptFilterName == "Identity":
                return (0, stream)
            else:
                # TODO: algorithm is cryptFilterName, specified in the /CF dictionary
                return (-1, "Decrypt not supported yet")


def dctDecode(stream, parameters):
    """
    Method to decode streams using a DCT technique based on the JPEG standard (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decoded PDF stream in case status = 0 or an error in case status = -1
    """
    decodedStream = ""
    try:
        from PIL import Image
        from io import BytesIO
    except:
        return (-1, "Python Imaging Library (PIL) not installed")
    # Quick implementation, assuming the library can detect the parameters
    try:
        im = Image.open(BytesIO(stream.encode("latin-1")))
        decodedStream = im.tobytes()
        return (0, decodedStream)
    except:
        return (-1, "Error decompressing image data")


def dctEncode(stream, parameters):
    """
    Method to encode streams using a DCT technique based on the JPEG standard (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encoded PDF stream in case status = 0 or an error in case status = -1
    """
    encodedStream = ""
    return (-1, "DctEncode not supported yet")


def jbig2Decode(stream, parameters):
    """
    Method to decode streams using the JBIG2 standard (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decoded PDF stream in case status = 0 or an error in case status = -1
    """
    decodedStream = ""
    return (-1, "Jbig2Decode not supported yet")


def jbig2Encode(stream, parameters):
    """
    Method to encode streams using the JBIG2 standard (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encoded PDF stream in case status = 0 or an error in case status = -1
    """
    encodedStream = ""
    return (-1, "Jbig2Encode not supported yet")


def jpxDecode(stream):
    """
    Method to decode streams using the JPEG2000 standard (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the decoded PDF stream in case status = 0 or an error in case status = -1
    """
    decodedStream = ""
    return (-1, "JpxDecode not supported yet")


def jpxEncode(stream):
    """
    Method to encode streams using the JPEG2000 standard (NOT IMPLEMENTED YET)

    @param stream: A PDF stream
    @return: A tuple (status,statusContent), where statusContent is the encoded PDF stream in case status = 0 or an error in case status = -1
    """
    encodedStream = ""
    return (-1, "JpxEncode not supported yet")
